//********************************************************************************************************************************
//*                                                                                                                              *
//*     APEv2 Library 1.0.22.28  3delite 2012-2022                                                                              *
//*     See APEv2 Library Readme.txt for details.                                                                                *
//*                                                                                                                              *
//* Licenses available for usage of this component:                                                                              *
//* Freeware License: 15                                                                                                        *
//* 	http://www.shareit.com/product.html?productid=300953174                                                                  *
//* Shareware License: 25                                                                                                       *
//* 	http://www.shareit.com/product.html?productid=300517941                                                                  *
//* Commercial License: 95                                                                                                      *
//* 	http://www.shareit.com/product.html?productid=300517939                                                                  *
//*                                                                                                                              *
//* Home page:                                                                                                                   *
//*                                                                                                                              *
//*     https://www.3delite.hu/Object%20Pascal%20Developer%20Resources/APEv2Library.html                                         *
//*                                                                                                                              *
//* This component is also available as a part of Tags Library:                                                                  *
//*                                                                                                                              *
//*     https://www.3delite.hu/Object%20Pascal%20Developer%20Resources/TagsLibrary.html                                          *
//*                                                                                                                              *
//* There is also an ID3v2 Library available at:                                                                                 *
//*                                                                                                                              *
//*     https://www.3delite.hu/Object%20Pascal%20Developer%20Resources/id3v2library.html                                         *
//*                                                                                                                              *
//* and also an MP4 Tag Library available at:                                                                                    *
//*                                                                                                                              *
//*     https://www.3delite.hu/Object%20Pascal%20Developer%20Resources/MP4TagLibrary.html                                        *
//*                                                                                                                              *
//* and also an Ogg Vorbis and Opus Tag Library available at:                                                                    *
//*                                                                                                                              *
//*     https://www.3delite.hu/Object%20Pascal%20Developer%20Resources/OpusTagLibrary.html                                       *
//*                                                                                                                              *
//* a Flac Tag Library available at:                                                                                             *
//*                                                                                                                              *
//*     https://www.3delite.hu/Object%20Pascal%20Developer%20Resources/FlacTagLibrary.html                                       *
//*                                                                                                                              *
//* a WAV Tag Library available at:                                                                                              *
//*                                                                                                                              *
//*     https://www.3delite.hu/Object%20Pascal%20Developer%20Resources/WAVTagLibrary.html                                        *
//*                                                                                                                              *
//* an WMA Tag Library available at:                                                                                             *
//*                                                                                                                              *
//*     https://www.3delite.hu/Object%20Pascal%20Developer%20Resources/WMATagLibrary.html                                        *
//*                                                                                                                              *
//* For other Delphi components see the home page:                                                                               *
//*                                                                                                                              *
//*     https://www.3delite.hu/                                                                                                  *
//*                                                                                                                              *
//* If you have any questions or enquiries please mail: 3delite@3delite.hu                                                       *
//*                                                                                                                              *
//* Good coding! :)                                                                                                              *
//* 3delite                                                                                                                      *
//********************************************************************************************************************************

unit APEv2Library;

{$MINENUMSIZE 4}

{$IFDEF FPC}
    {$MODE DELPHIUNICODE}{$H+}
    {$HINTS OFF}
    {$WARNINGS OFF}
{$ENDIF}

interface

Uses
    SysUtils,
    Classes,
    Generics.Collections;

Const
    APEv2LIBRARY_VERSION = $01002228;

Const
    APEV2LIBRARY_SUCCESS                    = 0;
    APEV2LIBRARY_ERROR                      = $FFFF;
    APEV2LIBRARY_ERROR_NO_TAG_FOUND         = 1;
    APEV2LIBRARY_ERROR_EMPTY_TAG            = 2;
    APEV2LIBRARY_ERROR_EMPTY_FRAMES         = 3;
    APEV2LIBRARY_ERROR_OPENING_FILE         = 4;
    APEV2LIBRARY_ERROR_READING_FILE         = 5;
    APEV2LIBRARY_ERROR_WRITING_FILE         = 6;
    APEV2LIBRARY_ERROR_CORRUPT              = 7;
    APEV2LIBRARY_ERROR_NOT_SUPPORTED_VERSION= 8;
    APEV2LIBRARY_ERROR_NOT_SUPPORTED_FORMAT = 9;

Const
    APEV2LIBRARY_DEFAULT_UPPERCASE_FIELD_NAMES = False;
    APEV2LIBRARY_DEFAULT_PRESERVE_FILE_DATETIME = False;

type
    TAPEv2ID = Array [1..8] of Byte;
    TID3v1ID = Array [0..2] of Byte;
    TID3LyricsTagIDEnd = Array [1..9] of Byte;
    TID3LyricsTagSize = Array [1..6] of Byte;

type
    TAPEv2Header = packed record
        Version: Cardinal;
        TagSize: Cardinal;
        ItemCount: Cardinal;
        Flags: Cardinal;
        Reserved: Int64;
    end;

type
    TAPEv2PictureFormat = (pfUnknown, pfJPEG, pfPNG, pfBMP, pfGIF);

type
    TAPEv2FrameFormat = (ffUnknown, ffText, ffBinary);

type
    TAPEv2Tag = class;

    TAPEv2Frame = class
    private
        function GetIndex: Integer;
    public
        Name: String;
        Format: TAPEv2FrameFormat;
        Stream: TMemoryStream;
        //Index: Integer;
        Parent: TAPEv2Tag;
        AlreadyParsed: Boolean;
        Constructor Create(_Parent: TAPEv2Tag);
        Destructor Destroy; override;
        function GetAsText: String;
        function GetAsTextANSI: String;
        function SetAsText(Text: String): Boolean;
        function GetAsList(var List: TStrings): Boolean;
        function SetAsList(List: TStrings): Boolean;
        function IsCoverArt: Boolean;
        procedure Clear;
        function Assign(APEv2Frame: TAPEv2Frame): Boolean;
        property Index: Integer read GetIndex;
    end;

    TAPEv2Tag = class
    private
        FFileSize: Int64;
    public
        FileName: String;
        Loaded: Boolean;
        Version: Cardinal;
        Flags: Cardinal;
        Size: Cardinal;
        Frames: TList<TAPEv2Frame>;
        UpperCaseFieldNamesToWrite: Boolean;
        PreserveFileDateTime: Boolean;
        Constructor Create;
        Destructor Destroy; override;
        function LoadFromFile(_FileName: String): Integer;
        function LoadFromStream(TagStream: TStream): Integer;
        function SaveToFile(_FileName: String): Integer;
        function SaveToStream(Stream: TStream): Integer;
        function AddFrame(const Name: String): TAPEv2Frame;
        function DeleteFrame(FrameIndex: Integer): Boolean;
        procedure DeleteAllFrames;
        procedure DeleteAllCoverArts;
        procedure Clear;
        function Count: Integer;
        function CoverArtCount: Integer;
        function GetFrameIndexOfCoverArtIndex(CoverArtIndex: Integer): Integer;
        function FrameExists(const Name: String): Integer; overload;
        function FrameTypeCount(const Name: String): Integer;
        function CalculateTotalFramesSize: Integer;
        function CalculateTagSize: Integer;
        function AddTextFrame(const Name: String; const Text: String): Integer;
        function AddBinaryFrame(const Name: String; BinaryStream: TStream; DataSize: Integer): Integer;
        procedure SetTextFrameText(const Name: String; const Text: String);
        procedure SetListFrameText(const Name: String; List: TStrings);
        function ReadFrameByNameAsText(const Name: String): String;
        function ReadFrameByNameAsList(const Name: String; var List: TStrings): Boolean;
        procedure RemoveEmptyFrames;
        function AddCoverArtFrame(const Name: String; PictureStream: TStream; Description: String): Integer;
        function SetCoverArtFrame(Index: Integer; PictureStream: TStream; Description: String): Boolean;
        function GetCoverArtFromFrame(Index: Integer; PictureStream: TStream; var PictureFormat: TAPEv2PictureFormat; var Description: String): Boolean;
        function GetCoverArtInfo(Index: Integer; var PictureFormat: TAPEv2PictureFormat; var Description: String): Boolean;
        function GetCoverArtInfoPointer(Index: Integer; var PictureFormat: TAPEv2PictureFormat; var Description: Pointer; var Data: Pointer; var DataSize: Int64): Boolean;
        function DeleteFrameByName(const Name: String): Boolean;
        function Assign(Source: TAPEv2Tag): Boolean;
        procedure ReSetAlreadyParsedState;
        property FileSize: Int64 read FFileSize;
    end;

    function APEv2ValidTag(TagStream: TStream): Boolean;

    function RemoveAPEv2FromFile(FileName: String; PreserveFileDateTime: Boolean = APEV2LIBRARY_DEFAULT_PRESERVE_FILE_DATETIME): Integer;
    function RemoveAPEv2FromStream(Stream: TStream): Integer;

    function APEv2TagErrorCode2String(ErrorCode: Integer): String;

var
    APEv2ID: TAPEv2ID;
    ID3v1ID: TID3v1ID;

    ID3LYRICSTAGIDSTART: Array[1..11] of Byte;
    ID3LYRICSTAGIDEND: Array[1..9] of Byte;

    APEv2TagLibraryDefaultUpperCaseFieldNamesToWrite: Boolean = APEV2LIBRARY_DEFAULT_UPPERCASE_FIELD_NAMES;
    APEv2TagLibraryDefaultPreserveFileDateTime: Boolean = APEV2LIBRARY_DEFAULT_PRESERVE_FILE_DATETIME;

implementation

{$IFDEF FPC}
function StrToUIntDef(Value: String; Default: QWord): Cardinal;
begin
     Result := StrToQWordDef(Value, Default);
end;

function SameText(const s1: WideString; const s2: WideString):Boolean;
begin
     Result := WideSameText(s1, s2);
end;
{$ENDIF}

function Bytes2MB(Bytes: Int64): String;
var
    KB: Extended;
    MB: Extended;
    GB: Extended;
    TB: Extended;
    PB: Extended;
    EB: Extended;
    ZB: Extended;
    YB: Extended;
    KBD: Extended;
    MBD: Extended;
    GBD: Extended;
    TBD: Extended;
    PBD: Extended;
    EBD: Extended;
    ZBD: Extended;
    YBD: Extended;
begin
    {
    if Bytes > 1048576 then begin
        if Bytes > 1073741824
            then Result := FloatToStrF((Bytes / 1073741824), ffFixed, 4, 2) + ' GB'
            else Result := FloatToStrF((Bytes / 1048576), ffFixed, 4, 2) + ' MB';
    end else Result := FloatToStrF((Bytes / 1024), ffFixed, 4, 2) + ' KB';
    if Bytes < 1024
 	    then Result := IntToStr(Bytes) + ' Byte';
    }

    KB := 1024.0 * 1024.0;
    MB := 1024.0 * 1024.0 * 1024.0;
    GB := 1024.0 * 1024.0 * 1024.0 * 1024.0;
    TB := 1024.0 * 1024.0 * 1024.0 * 1024.0 * 1024.0;
    PB := 1024.0 * 1024.0 * 1024.0 * 1024.0 * 1024.0 * 1024.0;
    EB := 1024.0 * 1024.0 * 1024.0 * 1024.0 * 1024.0 * 1024.0 * 1024.0;
    ZB := 1024.0 * 1024.0 * 1024.0 * 1024.0 * 1024.0 * 1024.0 * 1024.0 * 1024.0;
    YB := 1024.0 * 1024.0 * 1024.0 * 1024.0 * 1024.0 * 1024.0 * 1024.0 * 1024.0 * 1024.0;

    KBD := 1024.0;
    MBD := 1024.0 * 1024.0;
    GBD := 1024.0 * 1024.0 * 1024.0;
    TBD := 1024.0 * 1024.0 * 1024.0 * 1024.0;
    PBD := 1024.0 * 1024.0 * 1024.0 * 1024.0 * 1024.0;
    EBD := 1024.0 * 1024.0 * 1024.0 * 1024.0 * 1024.0 * 1024.0;
    ZBD := 1024.0 * 1024.0 * 1024.0 * 1024.0 * 1024.0 * 1024.0 * 1024.0;
    YBD := 1024.0 * 1024.0 * 1024.0 * 1024.0 * 1024.0 * 1024.0 * 1024.0 * 1024.0;

    if Bytes < 1024 then begin
        Result := IntToStr(Bytes) + ' Byte';
    end else begin
        if Bytes < KB then begin
            Result := FloatToStrF((Bytes / KBD), ffFixed, 4, 2) + ' KB';
        end else begin
            if Bytes < MB then begin
                Result := FloatToStrF((Bytes / MBD), ffFixed, 4, 2) + ' MB';
            end else begin
                if Bytes < GB then begin
                    Result := FloatToStrF((Bytes / GBD), ffFixed, 4, 2) + ' GB'
                end else begin
                    if Bytes < TB then begin
                        Result := FloatToStrF((Bytes / TBD), ffFixed, 4, 2) + ' TB'
                    end else begin
                        if Bytes < PB then begin
                            Result := FloatToStrF((Bytes / PBD), ffFixed, 4, 2) + ' PB'
                        end else begin
                            if Bytes < EB then begin
                                Result := FloatToStrF((Bytes / EBD), ffFixed, 4, 2) + ' EB'
                            end else begin
                                if Bytes < ZB then begin
                                    Result := FloatToStrF((Bytes / ZBD), ffFixed, 4, 2) + ' ZB'
                                end else begin
                                    if Bytes < YB then begin
                                        Result := FloatToStrF((Bytes / YBD), ffFixed, 4, 2) + ' YB'
                                    end else begin
                                        Result := FloatToStrF((Bytes / YBD), ffFixed, 4, 2) + ' YB'
                                    end;
                                end;
                            end;
                        end;
                    end;
                end;
            end;
        end;
    end;
end;

function GetLyricsTagSize(TagStream: TStream; ID3V1TAGSIZE: Integer): Integer;
var
    PreviousStreamPosition: Int64;
    LyricsTagIDEnd: TID3LyricsTagIDEnd;
    LyricsTagSize: TID3LyricsTagSize;
begin
    Result := 0;
    PreviousStreamPosition := TagStream.Position;
    try
        TagStream.Seek(- (ID3V1TAGSIZE + 9), soEnd);
        TagStream.Read(LyricsTagIDEnd, SizeOf(TID3LyricsTagIDEnd));
        if (LyricsTagIDEnd[1] = ID3LYRICSTAGIDEND[1])
        AND (LyricsTagIDEnd[2] = ID3LYRICSTAGIDEND[2])
        AND (LyricsTagIDEnd[3] = ID3LYRICSTAGIDEND[3])
        AND (LyricsTagIDEnd[4] = ID3LYRICSTAGIDEND[4])
        AND (LyricsTagIDEnd[5] = ID3LYRICSTAGIDEND[5])
        AND (LyricsTagIDEnd[6] = ID3LYRICSTAGIDEND[6])
        AND (LyricsTagIDEnd[7] = ID3LYRICSTAGIDEND[7])
        AND (LyricsTagIDEnd[8] = ID3LYRICSTAGIDEND[8])
        AND (LyricsTagIDEnd[9] = ID3LYRICSTAGIDEND[9])
        then begin
            TagStream.Seek(- (ID3V1TAGSIZE + 9 + 6), soEnd);
            TagStream.Read(LyricsTagSize, SizeOf(TID3LyricsTagSize));
            Result := StrToInt(Char(LyricsTagSize[1])
                 + Char(LyricsTagSize[2])
                 + Char(LyricsTagSize[3])
                 + Char(LyricsTagSize[4])
                 + Char(LyricsTagSize[5])
                 + Char(LyricsTagSize[6])
                 ) + 6 + 9;
        end;
    finally
        TagStream.Seek(PreviousStreamPosition, soBeginning);
    end;
end;

Constructor TAPEv2Frame.Create(_Parent: TAPEv2Tag);
begin
    Inherited Create;
    Self.Parent := _Parent;
    Name := '';
    //Flags := 0;
    Stream := TMemoryStream.Create;
    Format := ffUnknown;
end;

Destructor TAPEv2Frame.Destroy;
begin
    FreeAndNil(Stream);
    Inherited;
end;

function TAPEv2Frame.GetAsText: String;
var
    i: Integer;
    Data: Byte;
    Bytes: TBytes;
begin
    Result := '';
    if Format = ffBinary then begin
        Result := 'BINARY ' + Bytes2MB(Stream.Size);
    end else begin
        if Stream.Size < 1 then begin
            Exit;
        end;
        Stream.Seek(0, soBeginning);
        SetLength(Bytes, Stream.Size);
        for i := 0 to Stream.Size - 1 do begin
            Stream.Read(Data, 1);
            Bytes[i] := Data;
        end;
        Stream.Seek(0, soBeginning);
        try
            try
                Result := TEncoding.UTF8.GetString(Bytes);
                if Result = '' then begin
                    Result := TEncoding.ANSI.GetString(Bytes);
                end;
            except
                try
                    Result := TEncoding.ANSI.GetString(Bytes);
                except
                    //*
                end;
            end;
        except
            Result := 'BINARY ' + Bytes2MB(Stream.Size);
        end;
    end;
    Result := TrimRight(Result);
    AlreadyParsed := True;
end;

function TAPEv2Frame.GetAsTextANSI: String;
var
    i: Integer;
    Data: Byte;
    Bytes: TBytes;
begin
    Result := '';
    if Format = ffBinary then begin
        Result := 'BINARY ' + Bytes2MB(Stream.Size);
    end else begin
        if Stream.Size < 1 then begin
            Exit;
        end;
        Stream.Seek(0, soBeginning);
        SetLength(Bytes, Stream.Size);
        for i := 0 to Stream.Size - 1 do begin
            Stream.Read(Data, 1);
            Bytes[i] := Data;
        end;
        Stream.Seek(0, soBeginning);
        try
            Result := TEncoding.ANSI.GetString(Bytes);
        except
            Result := 'BINARY ' + Bytes2MB(Stream.Size);
        end;
    end;
    Result := TrimRight(Result);
    AlreadyParsed := True;
end;

function TAPEv2Frame.GetIndex: Integer;
begin
    Result := Parent.Frames.IndexOf(Self);
end;

function TAPEv2Frame.SetAsText(Text: String): Boolean;
var
    Bytes: TBytes;
begin
    Bytes := TEncoding.UTF8.GetBytes(Text);
    if Length(Bytes) = 0 then begin
        Bytes := TEncoding.ANSI.GetBytes(Text);
    end;
    Stream.Clear;
    Stream.Write(Bytes[0], Length(Bytes));
    Stream.Seek(0, soBeginning);
    Format := ffText;
    Result := True;
end;

function TAPEv2Frame.SetAsList(List: TStrings): Boolean;
var
    i: Integer;
    Data: Byte;
    NameBytes: TBytes;
    ValueBytes: TBytes;
begin
    //if Format <> ffText then begin
    //    Exit;
    //end;
    Stream.Clear;
    for i := 0 to List.Count - 1 do begin
        SetLength(NameBytes, 0);
        SetLength(ValueBytes, 0);
        NameBytes := TEncoding.UTF8.GetBytes(List.Names[i]);
        if Length(NameBytes) = 0 then begin
            NameBytes := TEncoding.ANSI.GetBytes(List.Names[i]);
        end;
        ValueBytes := TEncoding.UTF8.GetBytes(List.ValueFromIndex[i]);
        if Length(ValueBytes) = 0 then begin
            ValueBytes := TEncoding.ANSI.GetBytes(List.ValueFromIndex[i]);
        end;
        Stream.Write(NameBytes[0], Length(NameBytes));
        Data := $0D;
        Stream.Write(Data, 1);
        Data := $0A;
        Stream.Write(Data, 1);
        Stream.Write(ValueBytes[0], Length(ValueBytes));
        Data := $0D;
        Stream.Write(Data, 1);
        Data := $0A;
        Stream.Write(Data, 1);
    end;
    Stream.Seek(0, soBeginning);
    Format := ffText;
    Result := True;
end;

function TAPEv2Frame.GetAsList(var List: TStrings): Boolean;
var
    Data: Byte;
    Bytes: TBytes;
    NameStr: String;
    ValueStr: String;
    ByteCounter: Integer;
begin
    Result := False;
    List.Clear;
    if Format <> ffText then begin
        Exit;
    end;
    Stream.Seek(0, soBeginning);
    while Stream.Position < Stream.Size do begin
        SetLength(Bytes, 0);
        ByteCounter := 0;
        repeat
            Stream.Read(Data, 1);
            if Data = $0D then begin
                Stream.Read(Data, 1);
                if Data = $0A then begin
                    Break;
                end;
            end;
            SetLength(Bytes, Length(Bytes) + 1);
            Bytes[ByteCounter] := Data;
            Inc(ByteCounter);
        until Stream.Position >= Stream.Size;
        try
            NameStr := TEncoding.UTF8.GetString(Bytes);
            if NameStr = '' then begin
                NameStr := TEncoding.ANSI.GetString(Bytes);
            end;
        except
            try
                NameStr := TEncoding.ANSI.GetString(Bytes);
            except
                //*
            end;
        end;
        SetLength(Bytes, 0);
        ByteCounter := 0;
        repeat
            Stream.Read(Data, 1);
            if Data = $0D then begin
                Stream.Read(Data, 1);
                if Data = $0A then begin
                    Break;
                end;
            end;
            SetLength(Bytes, Length(Bytes) + 1);
            Bytes[ByteCounter] := Data;
            Inc(ByteCounter);
        until Stream.Position >= Stream.Size;
        try
            ValueStr := TEncoding.UTF8.GetString(Bytes);
            if ValueStr = '' then begin
                ValueStr := TEncoding.ANSI.GetString(Bytes);
            end;
        except
            try
                ValueStr := TEncoding.ANSI.GetString(Bytes);
            except
                //*
            end;
        end;
        List.Append(TrimRight(NameStr) + '=' + TrimRight(ValueStr));
        Result := True;
    end;
    Stream.Seek(0, soBeginning);
    AlreadyParsed := True;
end;

function TAPEv2Frame.IsCoverArt: Boolean;
var
    Bytes: TBytes;
    ByteCounter: Integer;
    DataSize: Cardinal;
    PPicture: PByte;
    Offset: Cardinal;
begin
    Result := False;
    if Format <> ffBinary then begin
        Exit;
    end;
    if SameText(Name, 'Cover Art (Front)') then begin
        Result := True;
        Exit;
    end;
    try
        try
            Offset := 0;
            DataSize := Stream.Size;
            PPicture := Stream.Memory;
            while Offset < DataSize do begin
                if Length(Bytes) = 0 then begin
                    ByteCounter := 0;
                    while (PPicture^ <> 0)
                    AND (Offset < DataSize)
                    do begin
                        SetLength(Bytes, Length(Bytes) + 1);
                        Bytes[ByteCounter] := PPicture^;
                        Inc(ByteCounter);
                        Inc(PPicture);
                        Inc(Offset);
                    end;
                end;
                //* JPEG
                if PPicture^ = $FF then begin
                    Inc(PPicture);
                    Inc(Offset);
                    if PPicture^ = $D8 then begin
                        Result := True;
                        Break;
                    end;
                end;
                //* PNG 89 50 4E 47
                if PPicture^ = $89 then begin
                    Inc(PPicture);
                    Inc(Offset);
                    if PPicture^ = $50 then begin
                        Inc(PPicture);
                        Inc(Offset);
                        if PPicture^ = $4E then begin
                            Inc(PPicture);
                            Inc(Offset);
                            if PPicture^ = $47 then begin
                                Result := True;
                                Break;
                            end;
                        end;
                    end;
                end;
                //* GIF 47 49 46 38
                if PPicture^ = $47 then begin
                    Inc(PPicture);
                    Inc(Offset);
                    if PPicture^ = $49 then begin
                        Inc(PPicture);
                        Inc(Offset);
                        if PPicture^ = $46 then begin
                            Inc(PPicture);
                            Inc(Offset);
                            if PPicture^ = $38 then begin
                                Result := True;
                                Break;
                            end;
                        end;
                    end;
                end;
                //* BMP 42 4D
                if PPicture^ = $42 then begin
                    Inc(PPicture);
                    Inc(Offset);
                    if PPicture^ = $4D then begin
                        Result := True;
                        Break;
                    end;
                end;
                Inc(PPicture);
                Inc(Offset);
            end;
        finally
            //*
        end;
    except
        Result := False;
    end;
end;

procedure TAPEv2Frame.Clear;
begin
    Format := ffUnknown;
    //Flags := 0;
    Stream.Clear;
end;

function TAPEv2Frame.Assign(APEv2Frame: TAPEv2Frame): Boolean;
begin
    Result := False;
    Self.Clear;
    if APEv2Frame <> nil then begin
        Format := APEv2Frame.Format;
        //Flags := APEv2Frame.Flags;
        APEv2Frame.Stream.Seek(0, soBeginning);
        Stream.CopyFrom(APEv2Frame.Stream, APEv2Frame.Stream.Size);
        Stream.Seek(0, soBeginning);
        APEv2Frame.Stream.Seek(0, soBeginning);
        Result := True;
    end;
end;

Constructor TAPEv2Tag.Create;
begin
    Inherited;
    Frames := TList<TAPEv2Frame>.Create;
    Clear;
    UpperCaseFieldNamesToWrite := APEv2TagLibraryDefaultUpperCaseFieldNamesToWrite;
    PreserveFileDateTime := APEv2TagLibraryDefaultPreserveFileDateTime;
end;

Destructor TAPEv2Tag.Destroy;
begin
    Clear;
    FreeAndNil(Frames);
    Inherited;
end;

procedure TAPEv2Tag.DeleteAllCoverArts;
var
    i: Integer;
begin
    for i := Count - 1 downto 0 do begin
        if Frames[i].IsCoverArt then begin
            DeleteFrame(i);
        end;
    end;
end;

function TAPEv2Tag.CoverArtCount: Integer;
var
    i: Integer;
begin
    Result := 0;
    for i := Count - 1 downto 0 do begin
        if Frames[i].IsCoverArt then begin
            Inc(Result);
        end;
    end;
end;

function TAPEv2Tag.GetFrameIndexOfCoverArtIndex(CoverArtIndex: Integer): Integer;
var
    i: Integer;
    Counter: Integer;
begin
    Result := - 1;
    Counter := 0;
    for i := 0 to Frames.Count - 1 do begin
        if Frames[i].IsCoverArt then begin
            if Counter = CoverArtIndex then begin
                Result := i;
                Break;
            end;
            Inc(Counter);
        end;
    end;
end;

procedure TAPEv2Tag.DeleteAllFrames;
var
    i: Integer;
begin
    for i := Frames.Count - 1 downto 0 do begin
        Frames[i].Free;
    end;
    Frames.Clear;
    Frames.Capacity := 256;
end;

function TAPEv2Tag.LoadFromStream(TagStream: TStream): Integer;
var
    PreviousPosition: Int64;
    ReadID3v1ID: TID3v1ID;
    APEv2Header: TAPEv2Header;
    i: Integer;
    DataSize: Cardinal;
    DataFlags: Cardinal;
    Data: Byte;
    FrameName: String;
    Bytes: TBytes;
    ByteCounter: Integer;
begin
    Result := APEV2LIBRARY_ERROR;
    Loaded := False;
    Clear;
    try
        PreviousPosition := TagStream.Position;
        try
            FFileSize := TagStream.Size;
            //if NOT APEv2ValidTag(TagStream) then begin
                TagStream.Seek(- 128, soEnd);
                TagStream.Read(ReadID3v1ID, 3);
                if (ReadID3v1ID[0] = ID3v1ID[0])
                AND (ReadID3v1ID[1] = ID3v1ID[1])
                AND (ReadID3v1ID[2] = ID3v1ID[2])
                then begin
                    TagStream.Seek(- 32 - 128 - GetLyricsTagSize(TagStream, 128), soEnd);
                end else begin
                    TagStream.Seek(- 32 - GetLyricsTagSize(TagStream, 0), soEnd);
                end;
                if NOT APEv2ValidTag(TagStream) then begin
                    Result := APEV2LIBRARY_ERROR_NO_TAG_FOUND;
                    Exit;
                end;
            //end;
            if TagStream.Read(APEv2Header, SizeOf(TAPEv2Header)) <> SizeOf(TAPEv2Header) then begin
                Result := APEV2LIBRARY_ERROR_NOT_SUPPORTED_FORMAT;
                Exit;
            end;
            Version := APEv2Header.Version;
            Flags := APEv2Header.Flags;
            Size := APEv2Header.TagSize;
            {
            if APEv2Header.Reserved <> 0 then begin
                Result := APEV2LIBRARY_ERROR_NOT_SUPPORTED_FORMAT;
                Exit;
            end;
            }
            if (APEv2Header.Version <> 2000) then begin
                if (APEv2Header.Version <> 1000) then begin
                    Result := APEV2LIBRARY_ERROR_NOT_SUPPORTED_VERSION;
                    Exit;
                end;
            end;
            {
            if APEv2Header.Flags AND NOT $80000001 <> 0 then begin
                Result := APEV2LIBRARY_ERROR_CORRUPT;
                Exit;
            end;
            }
            TagStream.Seek(- APEv2Header.TagSize, soCurrent);
            for i := 0 to APEv2Header.ItemCount - 1 do begin
                FrameName := '';
                TagStream.Read(DataSize, 4);
                TagStream.Read(DataFlags, 4);
                if DataFlags AND NOT $7 <> 0 then begin
                    Result := APEV2LIBRARY_ERROR_CORRUPT;
                    //Exit;
                end;
                if DataSize > Size then begin
                    Result := APEV2LIBRARY_ERROR_CORRUPT;
                    Exit;
                end;
                SetLength(Bytes, 0);
                ByteCounter := 0;
                repeat
                    TagStream.Read(Data, 1);
                    if Data <> 0 then begin
                        SetLength(Bytes, Length(Bytes) + 1);
                        Bytes[ByteCounter] := Data;
                        Inc(ByteCounter);
                    end;
                until Data = 0;
                try
                    FrameName := TEncoding.UTF8.GetString(Bytes);
                    if FrameName = '' then begin
                        FrameName := TEncoding.ANSI.GetString(Bytes);
                    end;
                except
                    try
                        FrameName := TEncoding.ANSI.GetString(Bytes);
                    except
                        Continue;
                    end;
                end;
                case (DataFlags and $6) SHR 1 of
                    0: begin
                        with AddFrame(FrameName) do begin
                            if DataSize > 0 then begin
                                Stream.CopyFrom(TagStream, DataSize);
                            end;
                            Format := ffText;
                        end;
                    end;
                    1: begin
                        with AddFrame(FrameName) do begin
                            if DataSize > 0 then begin
                                Stream.CopyFrom(TagStream, DataSize);
                            end;
                            Format := ffBinary;
                        end;
                    end;
                    //2: unsupported feature
                    3: begin
                        Result := APEV2LIBRARY_ERROR_CORRUPT;
                    end;
                end;
                Loaded := True;
            end;
        finally
            TagStream.Seek(PreviousPosition, soBeginning);
        end;
        //* Convert APEv1 ANSI text tags to APEv2 UTF8 format
        if APEv2Header.Version = 1000 then begin
            for i := 0 to Frames.Count - 1 do begin
                if Frames[i].Format = ffText then begin
                    Frames[i].SetAsText(Frames[i].GetAsTextANSI);
                end;
            end;
        end;
        Result := APEV2LIBRARY_SUCCESS;
    except
        if Result <> APEV2LIBRARY_ERROR_CORRUPT then begin
            Result := APEV2LIBRARY_ERROR;
        end;
    end;
end;

function TAPEv2Tag.LoadFromFile(_FileName: String): Integer;
var
    FileStream: TFileStream;
begin
    Clear;
    Loaded := False;
    if NOT FileExists(_FileName) then begin
        Result := APEV2LIBRARY_ERROR_OPENING_FILE;
        Exit;
    end;
    try
        FileStream := TFileStream.Create(_FileName, fmOpenRead OR fmShareDenyWrite);
    except
        Result := APEV2LIBRARY_ERROR_OPENING_FILE;
        Exit;
    end;
    try
        Result := LoadFromStream(FileStream);
        if (Result = APEV2LIBRARY_SUCCESS)
        OR (Result = APEV2LIBRARY_ERROR_NOT_SUPPORTED_VERSION)
        then begin
            Self.FileName := _FileName;
        end;
    finally
        FreeAndNil(FileStream);
    end;
end;

function TAPEv2Tag.AddFrame(const Name: String): TAPEv2Frame;
begin
    Result := nil;
    try
        if Frames.Count = Frames.Capacity then begin
            Frames.Capacity := Frames.Capacity + 256;
        end;
        Result := TAPEv2Frame.Create(Self);
        Result.Name := Name;
        Frames.Add(Result);
    except
        //*
    end;
end;

function TAPEv2Tag.DeleteFrame(FrameIndex: Integer): Boolean;
begin
    Result := False;
    if (FrameIndex >= Frames.Count)
    OR (FrameIndex < 0)
    then begin
        Exit;
    end;
    Frames[FrameIndex].Free;
    Frames.Delete(FrameIndex);
    Result := True;
end;

function TAPEv2Tag.FrameExists(const Name: String): Integer;
var
    i: Integer;
begin
    Result := -1;
    for i := 0 to Frames.Count - 1 do begin
        if Name = Frames[i].Name then begin
            Result := i;
            Break;
        end;
    end;
end;

function TAPEv2Tag.FrameTypeCount(const Name: String): Integer;
var
    i: Integer;
begin
    Result := 0;
    for i := 0 to Frames.Count - 1 do begin
        if UpperCase(Name) = UpperCase(Frames[i].Name) then begin
            Inc(Result);
        end;
    end;
end;

function TAPEv2Tag.SaveToFile(_FileName: String): Integer;
var
    FileDateTime: TDateTime;
    TagStream: TStream;
begin
    TagStream := nil;
    try
        try
            if PreserveFileDateTime then begin
                FileAge(_FileName, FileDateTime);
            end;
            RemoveEmptyFrames;
            if Frames.Count = 0 then begin
                Result := APEV2LIBRARY_ERROR_EMPTY_TAG;
                Exit;
            end;
            if CalculateTotalFramesSize = 0 then begin
                Result := APEV2LIBRARY_ERROR_EMPTY_FRAMES;
                Exit;
            end;
            if NOT FileExists(_FileName) then begin
                TagStream := TFileStream.Create(_FileName, fmCreate OR fmShareDenyWrite);
            end else begin
                TagStream := TFileStream.Create(_FileName, fmOpenReadWrite OR fmShareDenyWrite);
            end;
            Result := SaveToStream(TagStream);
        finally
            if Assigned(TagStream) then begin
                FreeAndNil(TagStream);
            end;
            if PreserveFileDateTime then begin
                FileSetDate(_FileName, DateTimeToFileDate(FileDateTime));
            end;
        end;
    except
        Result := APEV2LIBRARY_ERROR;
    end;
end;

function TAPEv2Tag.SaveToStream(Stream: TStream): Integer;
var
    ReadID3v1ID: TID3v1ID;
    APEv2Header: TAPEv2Header;
    ID3v1Stream: TMemoryStream;
    APEv2TagExists: Boolean;
    i: Integer;
    Bytes: TBytes;
    FrameSize: Cardinal;
    FrameFlags: Cardinal;
    Data: Byte;
    LyricsTagSize: Integer;
begin
    APEv2TagExists := False;
    try
        try
            RemoveEmptyFrames;
            if Frames.Count = 0 then begin
                Result := APEV2LIBRARY_ERROR_EMPTY_TAG;
                Exit;
            end;
            if CalculateTotalFramesSize = 0 then begin
                Result := APEV2LIBRARY_ERROR_EMPTY_FRAMES;
                Exit;
            end;
            ID3v1Stream := TMemoryStream.Create;
            if Stream.Size >= 128 then begin
                Stream.Seek(- 128, soEnd);
                Stream.Read(ReadID3v1ID, 3);
            end;
            if (ReadID3v1ID[0] = ID3v1ID[0])
            AND (ReadID3v1ID[1] = ID3v1ID[1])
            AND (ReadID3v1ID[2] = ID3v1ID[2])
            then begin
                LyricsTagSize := GetLyricsTagSize(Stream, 128);
                Stream.Seek(- 128 - LyricsTagSize, soEnd);
                ID3v1Stream.CopyFrom(Stream, 128 + LyricsTagSize);
                ID3v1Stream.Seek(0, soBeginning);
                Stream.Seek(- 32 - 128 - LyricsTagSize, soEnd);
            end else begin
                LyricsTagSize := GetLyricsTagSize(Stream, 0);
                if LyricsTagSize > 0 then begin
                    Stream.Seek(- LyricsTagSize, soEnd);
                    ID3v1Stream.CopyFrom(Stream, LyricsTagSize);
                    ID3v1Stream.Seek(0, soBeginning);
                end;
                Stream.Seek(- 32 - LyricsTagSize, soEnd);
            end;
            if APEv2ValidTag(Stream) then begin
                APEv2TagExists := True;
            end;
            if NOT APEv2TagExists then begin
                //if ID3v1Stream.Size <> 0 then begin
                    Stream.Seek(- ID3v1Stream.Size, soEnd);
                //end else begin
                //    Stream.Seek(0, soEnd);
                //end;
            end else begin
                //if ID3v1Stream.Size <> 0 then begin
                    Stream.Seek(- 24 - ID3v1Stream.Size, soEnd);
                //end else begin
                //    Stream.Seek(- 24, soEnd);
                //end;
                if Stream.Read(APEv2Header, SizeOf(TAPEv2Header)) <> SizeOf(TAPEv2Header) then begin
                    Result := APEV2LIBRARY_ERROR_NOT_SUPPORTED_FORMAT;
                    Exit;
                end;
                {
                if APEv2Header.Reserved <> 0 then begin
                    Result := APEV2LIBRARY_ERROR_NOT_SUPPORTED_FORMAT;
                    Exit;
                end;
                }
                if (APEv2Header.Version <> 2000) then begin
                    if (APEv2Header.Version <> 1000) then begin
                        Result := APEV2LIBRARY_ERROR_NOT_SUPPORTED_VERSION;
                        Exit;
                    end;
                end;
                {
                if APEv2Header.Flags AND NOT $80000001 <> 0 then begin
                    Result := APEV2LIBRARY_ERROR_CORRUPT;
                    Exit;
                end;
                }
                Stream.Seek(- APEv2Header.TagSize, soCurrent);
                Stream.Seek(- 32, soCurrent);
            end;
            //* Do the write
            APEv2Header.Version := 2000;
            APEv2Header.TagSize := CalculateTagSize;
            APEv2Header.ItemCount := Frames.Count;
            APEv2Header.Flags := $a0000000;
            APEv2Header.Reserved := 0;
            Stream.Write(APEv2ID, 8);
            Stream.Write(APEv2Header, SizeOf(TAPEv2Header));
            for i := 0 to Frames.Count - 1 do begin
                if UpperCaseFieldNamesToWrite then begin
                    Bytes := TEncoding.UTF8.GetBytes(UpperCase(Frames[i].Name));
                    if Length(Bytes) = 0 then begin
                        Bytes := TEncoding.ANSI.GetBytes(UpperCase(Frames[i].Name));
                    end;
                end else begin
                    Bytes := TEncoding.UTF8.GetBytes(Frames[i].Name);
                    if Length(Bytes) = 0 then begin
                        Bytes := TEncoding.ANSI.GetBytes(Frames[i].Name);
                    end;
                end;
                case Frames[i].Format of
                    ffText: begin
                        FrameFlags := 0;
                    end;
                    ffBinary: begin
                        FrameFlags := $2;
                    end;
                end;
                FrameSize := Frames[i].Stream.Size;
                Stream.Write(FrameSize, 4);
                Stream.Write(FrameFlags, 4);
                Stream.Write(Bytes[0], Length(Bytes));
                Data := $0;
                Stream.Write(Data, 1);
                Frames[i].Stream.Seek(0, soBeginning);
                Stream.CopyFrom(Frames[i].Stream, Frames[i].Stream.Size);
            end;
            Stream.Write(APEv2ID, 8);
            APEv2Header.Flags := $a0000000 AND NOT $20000000;
            Stream.Write(APEv2Header, SizeOf(TAPEv2Header));
            Stream.CopyFrom(ID3v1Stream, ID3v1Stream.Size);
            TFileStream(Stream).Size := Stream.Position;
            Result := APEV2LIBRARY_SUCCESS;
        finally
            if Assigned(ID3v1Stream) then begin
                FreeAndNil(ID3v1Stream);
            end;
        end;
    except
        Result := APEV2LIBRARY_ERROR;
    end;
end;

function TAPEv2Tag.CalculateTagSize: Integer;
var
    TotalTagSize: Integer;
begin
    TotalTagSize := CalculateTotalFramesSize;
    TotalTagSize := TotalTagSize + 8 + SizeOf(TAPEv2Header);
    Result := TotalTagSize;
end;

function TAPEv2Tag.CalculateTotalFramesSize: Integer;
var
    TotalFramesSize: Integer;
    i: Integer;
    Bytes: TBytes;
begin
    TotalFramesSize := 0;
    for i := 0 to Frames.Count - 1 do begin
        Bytes := TEncoding.UTF8.GetBytes(Frames[i].Name);
        if Length(Bytes) = 0 then begin
            Bytes := TEncoding.ANSI.GetBytes(Frames[i].Name);
        end;
        TotalFramesSize := TotalFramesSize + Frames[i].Stream.Size + Length(Bytes) + 1 + 4 + 4;
    end;
    Result := TotalFramesSize;
end;

procedure TAPEv2Tag.Clear;
begin
    DeleteAllFrames;
    FileName := '';
    Loaded := False;
    Version := 0;
    Flags := 0;
    Size := 0;
    FFileSize := 0;
end;

function TAPEv2Tag.Count: Integer;
begin
    Result := Frames.Count;
end;

function TAPEv2Tag.AddTextFrame(const Name: String; const Text: String): Integer;
begin
    with AddFrame(Name) do begin
        SetAsText(Text);
        Result := Index;
    end;
end;

function TAPEv2Tag.AddBinaryFrame(const Name: String; BinaryStream: TStream; DataSize: Integer): Integer;
var
    PreviousPosition: Int64;
begin
    with AddFrame(Name) do begin
        PreviousPosition := BinaryStream.Position;
        Stream.CopyFrom(BinaryStream, DataSize);
        Format := ffBinary;
        BinaryStream.Seek(PreviousPosition, soBeginning);
        Result := Index;
    end;
end;

procedure TAPEv2Tag.SetTextFrameText(const Name: String; const Text: String);
var
    i: Integer;
    l: Integer;
begin
    i := 0;
    l := Frames.Count;
    while (i < l)
    AND (UpperCase(Frames[i].Name) <> UpperCase(Name))
    do begin
        inc(i);
    end;
    if i = l then begin
        AddTextFrame(Name, Text);
    end else begin
        Frames[i].SetAsText(Text);
    end;
end;

procedure TAPEv2Tag.SetListFrameText(const Name: String; List: TStrings);
var
    i: Integer;
    l: Integer;
begin
    i := 0;
    l := Frames.Count;
    while (i < l)
    AND (UpperCase(Frames[i].Name) <> UpperCase(Name))
    do begin
        inc(i);
    end;
    if i = l then begin
        AddFrame(Name).SetAsList(List);
    end else begin
        Frames[i].SetAsList(List);
    end;
end;

function TAPEv2Tag.ReadFrameByNameAsText(const Name: String): String;
var
    i: Integer;
    l: Integer;
begin
    Result := '';
    l := Frames.Count;
    i := 0;
    while (i <> l)
    AND (UpperCase(Frames[i].Name) <> UpperCase(Name))
    do begin
        inc(i);
    end;
    if i = l then begin
        Result := '';
    end else begin
        if Frames[i].Format = ffText then begin
            Result := Frames[i].GetAsText;
        end;
    end;
end;

function TAPEv2Tag.ReadFrameByNameAsList(const Name: String; var List: TStrings): Boolean;
var
    i: Integer;
    l: Integer;
begin
    Result := False;
    l := Frames.Count;
    i := 0;
    while (i <> l)
    AND (UpperCase(Frames[i].Name) <> UpperCase(Name))
    do begin
        inc(i);
    end;
    if i = l then begin
        Result := False;
    end else begin
        if Frames[i].Format = ffText then begin
            Result := Frames[i].GetAsList(List);
        end;
    end;
end;

procedure TAPEv2Tag.RemoveEmptyFrames;
var
    i: Integer;
begin
    for i := Frames.Count - 1 downto 0 do begin
        if Frames[i].Stream.Size = 0 then begin
            DeleteFrame(i);
        end;
    end;
end;

procedure TAPEv2Tag.ReSetAlreadyParsedState;
var
    i: Integer;
begin
    for i := 0 to Frames.Count - 1 do begin
        Frames[i].AlreadyParsed := False;
    end;
end;

function TAPEv2Tag.AddCoverArtFrame(const Name: String; PictureStream: TStream; Description: String): Integer;
var
    Data: TMemoryStream;
    ZeroByte: Byte;
    Bytes: TBytes;
begin
    ZeroByte := 0;
    Data := TMemoryStream.Create;
    try
        Bytes := TEncoding.UTF8.GetBytes(Description);
        if Length(Bytes) = 0 then begin
            Bytes := TEncoding.ANSI.GetBytes(Description);
        end;
        Data.Write(Bytes[0], Length(Bytes));
        Data.Write(ZeroByte, 1);
        if Assigned(PictureStream) then begin
            PictureStream.Seek(0, soBeginning);
            Data.CopyFrom(PictureStream, PictureStream.Size);
            PictureStream.Seek(0, soBeginning);
        end;
        Data.Seek(0, soBeginning);
        Result := AddBinaryFrame(Name, Data, Data.Size);
    finally
        FreeAndNil(Data);
    end;
end;

function TAPEv2Tag.SetCoverArtFrame(Index: Integer; PictureStream: TStream; Description: String): Boolean;
var
    Data: TMemoryStream;
    ZeroByte: Byte;
    Bytes: TBytes;
begin
    Result := False;
    if (Index >= Frames.Count)
    OR (Index < 0)
    then begin
        Exit;
    end;
    ZeroByte := 0;
    Data := TMemoryStream.Create;
    try
        Bytes := TEncoding.UTF8.GetBytes(Description);
        if Length(Bytes) = 0 then begin
            Bytes := TEncoding.ANSI.GetBytes(Description);
        end;
        Data.Write(Bytes[0], Length(Bytes));
        Data.Write(ZeroByte, 1);
        PictureStream.Seek(0, soBeginning);
        Data.CopyFrom(PictureStream, PictureStream.Size);
        Data.Seek(0, soBeginning);
        Frames[Index].Stream.Clear;
        Frames[Index].Stream.CopyFrom(Data, Data.Size);
        PictureStream.Seek(0, soBeginning);
        Result := True;
    finally
        FreeAndNil(Data);
    end;
end;

function TAPEv2Tag.GetCoverArtFromFrame(Index: Integer; PictureStream: TStream; var PictureFormat: TAPEv2PictureFormat; var Description: String): Boolean;
var
    DataSize: Cardinal;
    PPicture: PByte;
    Offset: Cardinal;
    Bytes: TBytes;
    ByteCounter: Integer;
begin
    Result := False;
    Description := '';
    PictureFormat := pfUnknown;
    if (Index >= Frames.Count)
    OR (Index < 0)
    then begin
        Exit;
    end;
    try
        Offset := 0;
        DataSize := Frames[Index].Stream.Size;
        PPicture := Frames[Index].Stream.Memory;
        while Offset < DataSize do begin
            if Length(Bytes) = 0 then begin
                ByteCounter := 0;
                while (PPicture^ <> 0)
                AND (Offset < DataSize)
                do begin
                    SetLength(Bytes, Length(Bytes) + 1);
                    Bytes[ByteCounter] := PPicture^;
                    Inc(ByteCounter);
                    Inc(PPicture);
                    Inc(Offset);
                end;
            end;
            //* JPEG
            if PPicture^ = $FF then begin
                Inc(PPicture);
                Inc(Offset);
                if PPicture^ = $D8 then begin
                    PictureFormat := pfJPEG;
                    Dec(PPicture);
                    Dec(Offset);
                    Result := True;
                    Break;
                end;
            end;
            //* PNG 89 50 4E 47 0D 0A 1A 0A
            if PPicture^ = $89 then begin
                Inc(PPicture);
                Inc(Offset);
                if PPicture^ = $50 then begin
                    Inc(PPicture);
                    Inc(Offset);
                    if PPicture^ = $4E then begin
                        Inc(PPicture);
                        Inc(Offset);
                        if PPicture^ = $47 then begin
                            PictureFormat := pfPNG;
                            Dec(PPicture, 3);
                            Dec(Offset, 3);
                            Result := True;
                            Break;
                        end;
                    end;
                end;
            end;
            //* GIF 47 49 46 38
            if PPicture^ = $47 then begin
                Inc(PPicture);
                Inc(Offset);
                if PPicture^ = $49 then begin
                    Inc(PPicture);
                    Inc(Offset);
                    if PPicture^ = $46 then begin
                        Inc(PPicture);
                        Inc(Offset);
                        if PPicture^ = $38 then begin
                            PictureFormat := pfGIF;
                            Dec(PPicture, 3);
                            Dec(Offset, 3);
                            Result := True;
                            Break;
                        end;
                    end;
                end;
            end;
            //* BMP 42 4D
            if PPicture^ = $42 then begin
                Inc(PPicture);
                Inc(Offset);
                if PPicture^ = $4D then begin
                    PictureFormat := pfBMP;
                    Dec(PPicture);
                    Dec(Offset);
                    Result := True;
                    Break;
                end;
            end;
            Inc(PPicture);
            Inc(Offset);
        end;
        PictureStream.Write(PPicture^, DataSize - Offset);
        PictureStream.Seek(0, soBeginning);
        try
            Description := TEncoding.UTF8.GetString(Bytes);
            if Description = '' then begin
                Description := TEncoding.ANSI.GetString(Bytes);
            end;
        except
            try
                Description := TEncoding.ANSI.GetString(Bytes);
            except
                //*
            end;
        end;
        if Result then begin
            //*
        end else begin
            Description := '';
        end;
    except
        Result := False;
    end;
end;

function TAPEv2Tag.GetCoverArtInfo(Index: Integer; var PictureFormat: TAPEv2PictureFormat; var Description: String): Boolean;
var
    BinaryStream: TMemoryStream;
    DataSize: Cardinal;
    PPicture: PByte;
    Offset: Cardinal;
    Bytes: TBytes;
    ByteCounter: Integer;
begin
    Result := False;
    Description := '';
    PictureFormat := pfUnknown;
    if (Index >= Frames.Count)
    OR (Index < 0)
    then begin
        Exit;
    end;
    try
        BinaryStream := TMemoryStream.Create;
        try
            Offset := 0;
            DataSize := Frames[Index].Stream.Size;
            PPicture := Frames[Index].Stream.Memory;
            while Offset < DataSize do begin
                if Length(Bytes) = 0 then begin
                    ByteCounter := 0;
                    while (PPicture^ <> 0)
                    AND (Offset < DataSize)
                    do begin
                        SetLength(Bytes, Length(Bytes) + 1);
                        Bytes[ByteCounter] := PPicture^;
                        Inc(ByteCounter);
                        Inc(PPicture);
                        Inc(Offset);
                    end;
                end;
                //* JPEG
                if PPicture^ = $FF then begin
                    Inc(PPicture);
                    Inc(Offset);
                    if PPicture^ = $D8 then begin
                        PictureFormat := pfJPEG;
                        Dec(PPicture);
                        Dec(Offset);
                        Result := True;
                        Break;
                    end;
                end;
                //* PNG 89 50 4E 47 0D 0A 1A 0A
                if PPicture^ = $89 then begin
                    Inc(PPicture);
                    Inc(Offset);
                    if PPicture^ = $50 then begin
                        Inc(PPicture);
                        Inc(Offset);
                        if PPicture^ = $4E then begin
                            Inc(PPicture);
                            Inc(Offset);
                            if PPicture^ = $47 then begin
                                PictureFormat := pfPNG;
                                Dec(PPicture, 3);
                                Dec(Offset, 3);
                                Result := True;
                                Break;
                            end;
                        end;
                    end;
                end;
                //* GIF 47 49 46 38
                if PPicture^ = $47 then begin
                    Inc(PPicture);
                    Inc(Offset);
                    if PPicture^ = $49 then begin
                        Inc(PPicture);
                        Inc(Offset);
                        if PPicture^ = $46 then begin
                            Inc(PPicture);
                            Inc(Offset);
                            if PPicture^ = $38 then begin
                                PictureFormat := pfGIF;
                                Dec(PPicture, 3);
                                Dec(Offset, 3);
                                Result := True;
                                Break;
                            end;
                        end;
                    end;
                end;
                //* BMP 42 4D
                if PPicture^ = $42 then begin
                    Inc(PPicture);
                    Inc(Offset);
                    if PPicture^ = $4D then begin
                        PictureFormat := pfBMP;
                        Dec(PPicture);
                        Dec(Offset);
                        Result := True;
                        Break;
                    end;
                end;
                Inc(PPicture);
                Inc(Offset);
            end;
            BinaryStream.Write(PPicture^, DataSize - Offset);
            BinaryStream.Seek(0, soBeginning);
            try
                Description := TEncoding.UTF8.GetString(Bytes);
                if Description = '' then begin
                    Description := TEncoding.ANSI.GetString(Bytes);
                end;
            except
                try
                    Description := TEncoding.ANSI.GetString(Bytes);
                except
                    //*
                end;
            end;
            if Result then begin
                //*
            end else begin
                Description := '';
            end;
        finally
            FreeAndNil(BinaryStream);
        end;
    except
        Result := False;
    end;
end;

function TAPEv2Tag.GetCoverArtInfoPointer(Index: Integer; var PictureFormat: TAPEv2PictureFormat; var Description: Pointer; var Data: Pointer; var DataSize: Int64): Boolean;
var
    PPicture: PByte;
    Offset: Cardinal;
    Bytes: TBytes;
    ByteCounter: Integer;
begin
    Result := False;
    PictureFormat := pfUnknown;
    if (Index >= Frames.Count)
    OR (Index < 0)
    then begin
        Exit;
    end;
    try
        Offset := 0;
        DataSize := Frames[Index].Stream.Size;
        PPicture := Frames[Index].Stream.Memory;
        Description := Frames[Index].Stream.Memory;
        while Offset < DataSize do begin
            if Length(Bytes) = 0 then begin
                ByteCounter := 0;
                while (PPicture^ <> 0)
                AND (Offset < DataSize)
                do begin
                    SetLength(Bytes, Length(Bytes) + 1);
                    Bytes[ByteCounter] := PPicture^;
                    Inc(ByteCounter);
                    Inc(PPicture);
                    Inc(Offset);
                end;
            end;
            //* JPEG
            if PPicture^ = $FF then begin
                Inc(PPicture);
                Inc(Offset);
                if PPicture^ = $D8 then begin
                    PictureFormat := pfJPEG;
                    Dec(PPicture);
                    Dec(Offset);
                    Result := True;
                    Break;
                end;
            end;
            //* PNG 89 50 4E 47 0D 0A 1A 0A
            if PPicture^ = $89 then begin
                Inc(PPicture);
                Inc(Offset);
                if PPicture^ = $50 then begin
                    Inc(PPicture);
                    Inc(Offset);
                    if PPicture^ = $4E then begin
                        Inc(PPicture);
                        Inc(Offset);
                        if PPicture^ = $47 then begin
                            PictureFormat := pfPNG;
                            Dec(PPicture, 3);
                            Dec(Offset, 3);
                            Result := True;
                            Break;
                        end;
                    end;
                end;
            end;
            //* GIF 47 49 46 38
            if PPicture^ = $47 then begin
                Inc(PPicture);
                Inc(Offset);
                if PPicture^ = $49 then begin
                    Inc(PPicture);
                    Inc(Offset);
                    if PPicture^ = $46 then begin
                        Inc(PPicture);
                        Inc(Offset);
                        if PPicture^ = $38 then begin
                            PictureFormat := pfGIF;
                            Dec(PPicture, 3);
                            Dec(Offset, 3);
                            Result := True;
                            Break;
                        end;
                    end;
                end;
            end;
            //* BMP 42 4D
            if PPicture^ = $42 then begin
                Inc(PPicture);
                Inc(Offset);
                if PPicture^ = $4D then begin
                    PictureFormat := pfBMP;
                    Dec(PPicture);
                    Dec(Offset);
                    Result := True;
                    Break;
                end;
            end;
            Inc(PPicture);
            Inc(Offset);
        end;
        Data := PPicture;
        DataSize := DataSize - Offset;
    except
        Result := False;
    end;
end;

function TAPEv2Tag.DeleteFrameByName(const Name: String): Boolean;
var
    i: Integer;
    l: Integer;
begin
    l := Frames.Count;
    i := 0;
    while (i <> l) and (UpperCase(Frames[i].Name) <> UpperCase(Name)) do begin
        inc(i);
    end;
    if i = l then begin
        Result := False;
        Exit;
    end;
    Frames[i].Free;
    Frames.Delete(i);
    Result := True;
end;

function TAPEv2Tag.Assign(Source: TAPEv2Tag): Boolean;
var
    i: Integer;
begin
    Clear;
    try
        FileName := Source.FileName;
        Loaded := Source.Loaded;
        Version := Source.Version;
        Flags := Source.Flags;
        Size := Source.Size;
        for i := 0 to Source.Frames.Count - 1 do begin
            case Source.Frames[i].Format of
                ffText: begin
                    SetTextFrameText(Source.Frames[i].Name, Source.Frames[i].GetAsText);
                end;
                ffBinary: begin
                    Source.Frames[i].Stream.Seek(0, soBeginning);
                    AddBinaryFrame(Source.Frames[i].Name, Source.Frames[i].Stream, Source.Frames[i].Stream.Size);
                end;
            end;
        end;
        Result := True;
    except
        Result := False;
    end;
end;

function APEv2ValidTag(TagStream: TStream): Boolean;
var
    Identification: TAPEv2ID;
begin
    Result := False;
    try
        FillChar(Identification, SizeOf(Identification), 0);
        TagStream.Read(Identification[1], 8);
        if (Identification[1] = APEv2ID[1])
        AND (Identification[2] = APEv2ID[2])
        AND (Identification[3] = APEv2ID[3])
        AND (Identification[4] = APEv2ID[4])
        AND (Identification[5] = APEv2ID[5])
        AND (Identification[6] = APEv2ID[6])
        AND (Identification[7] = APEv2ID[7])
        AND (Identification[8] = APEv2ID[8])
        then begin
            Result := True;
        end;
    except
        //*
    end;
end;

function RemoveAPEv2FromFile(FileName: String; PreserveFileDateTime: Boolean = APEV2LIBRARY_DEFAULT_PRESERVE_FILE_DATETIME): Integer;
var
    FileDateTime: TDateTime;
    TagStream: TStream;
    ReadID3v1ID: TID3v1ID;
    APEv2Header: TAPEv2Header;
    ID3v1Stream: TMemoryStream;
begin
    TagStream := nil;
    try
        try
            if PreserveFileDateTime then begin
                FileAge(FileName, FileDateTime);
            end;
            if NOT FileExists(FileName) then begin
                TagStream := TFileStream.Create(FileName, fmCreate OR fmShareDenyWrite);
            end else begin
                TagStream := TFileStream.Create(FileName, fmOpenReadWrite OR fmShareDenyWrite);
            end;
            ID3v1Stream := TMemoryStream.Create;
            if TagStream.Size >= 128 then begin
                TagStream.Seek(- 128, soEnd);
                TagStream.Read(ReadID3v1ID, 3);
            end;
            if (ReadID3v1ID[0] = ID3v1ID[0])
            AND (ReadID3v1ID[1] = ID3v1ID[1])
            AND (ReadID3v1ID[2] = ID3v1ID[2])
            then begin
                TagStream.Seek(- 128, soEnd);
                ID3v1Stream.CopyFrom(TagStream, 128);
                ID3v1Stream.Seek(0, soBeginning);
                TagStream.Seek(- 32 - 128, soEnd);
            end else begin
                TagStream.Seek(- 32, soEnd);
            end;
            if NOT APEv2ValidTag(TagStream) then begin
                Result := APEV2LIBRARY_ERROR_NO_TAG_FOUND;
                Exit;
            end;
            if ID3v1Stream.Size <> 0 then begin
                TagStream.Seek(- 24 - 128, soEnd);
            end else begin
                TagStream.Seek(- 24, soEnd);
            end;
            if TagStream.Read(APEv2Header, SizeOf(TAPEv2Header)) <> SizeOf(TAPEv2Header) then begin
                Result := APEV2LIBRARY_ERROR_NOT_SUPPORTED_FORMAT;
                Exit;
            end;
            if APEv2Header.Reserved <> 0 then begin
                Result := APEV2LIBRARY_ERROR_NOT_SUPPORTED_FORMAT;
                Exit;
            end;
            if (APEv2Header.Version <> 2000) then begin
                if (APEv2Header.Version <> 1000) then begin
                    Result := APEV2LIBRARY_ERROR_NOT_SUPPORTED_VERSION;
                    Exit;
                end;
            end;
            {
            if APEv2Header.Flags AND NOT $80000001 <> 0 then begin
                Result := APEV2LIBRARY_ERROR_CORRUPT;
                Exit;
            end;
            }
            TagStream.Seek(- APEv2Header.TagSize, soCurrent);
            TagStream.Seek(- 32, soCurrent);
            TFileStream(TagStream).Size := TagStream.Position;
            TagStream.CopyFrom(ID3v1Stream, ID3v1Stream.Size);
            Result := APEV2LIBRARY_SUCCESS;
        finally
            if Assigned(TagStream) then begin
                FreeAndNil(TagStream);
            end;
            if Assigned(ID3v1Stream) then begin
                FreeAndNil(ID3v1Stream);
            end;
            if PreserveFileDateTime then begin
                FileSetDate(FileName, DateTimeToFileDate(FileDateTime));
            end;
        end;
    except
        Result := APEV2LIBRARY_ERROR;
    end;
end;

function RemoveAPEv2FromStream(Stream: TStream): Integer;
var
    ReadID3v1ID: TID3v1ID;
    APEv2Header: TAPEv2Header;
    ID3v1Stream: TMemoryStream;
begin
    if Stream.Size = 0 then begin
        Result := APEV2LIBRARY_SUCCESS;
        Exit;
    end;
    try
        try
            ID3v1Stream := TMemoryStream.Create;
            if Stream.Size >= 128 then begin
                Stream.Seek(- 128, soEnd);
                Stream.Read(ReadID3v1ID, 3);
            end;
            if (ReadID3v1ID[0] = ID3v1ID[0])
            AND (ReadID3v1ID[1] = ID3v1ID[1])
            AND (ReadID3v1ID[2] = ID3v1ID[2])
            then begin
                Stream.Seek(- 128, soEnd);
                ID3v1Stream.CopyFrom(Stream, 128);
                ID3v1Stream.Seek(0, soBeginning);
                Stream.Seek(- 32 - 128, soEnd);
            end else begin
                Stream.Seek(- 32, soEnd);
            end;
            if NOT APEv2ValidTag(Stream) then begin
                Result := APEV2LIBRARY_ERROR_NO_TAG_FOUND;
                Exit;
            end;
            if ID3v1Stream.Size <> 0 then begin
                Stream.Seek(- 24 - 128, soEnd);
            end else begin
                Stream.Seek(- 24, soEnd);
            end;
            if Stream.Read(APEv2Header, SizeOf(TAPEv2Header)) <> SizeOf(TAPEv2Header) then begin
                Result := APEV2LIBRARY_ERROR_NOT_SUPPORTED_FORMAT;
                Exit;
            end;
            if APEv2Header.Reserved <> 0 then begin
                Result := APEV2LIBRARY_ERROR_NOT_SUPPORTED_FORMAT;
                Exit;
            end;
            if (APEv2Header.Version <> 2000) then begin
                if (APEv2Header.Version <> 1000) then begin
                    Result := APEV2LIBRARY_ERROR_NOT_SUPPORTED_VERSION;
                    Exit;
                end;
            end;
            {
            if APEv2Header.Flags AND NOT $80000001 <> 0 then begin
                Result := APEV2LIBRARY_ERROR_CORRUPT;
                Exit;
            end;
            }
            Stream.Seek(- APEv2Header.TagSize, soCurrent);
            Stream.Seek(- 32, soCurrent);
            Stream.Size := Stream.Position;
            Stream.CopyFrom(ID3v1Stream, ID3v1Stream.Size);
            Stream.Seek(0, soBeginning);
            Result := APEV2LIBRARY_SUCCESS;
        finally
            if Assigned(ID3v1Stream) then begin
                FreeAndNil(ID3v1Stream);
            end;
        end;
    except
        Result := APEV2LIBRARY_ERROR;
    end;
end;

function APEv2TagErrorCode2String(ErrorCode: Integer): String;
begin
    Result := 'Unknown error code.';
    case ErrorCode of
        APEV2LIBRARY_SUCCESS: Result := 'Success.';
        APEV2LIBRARY_ERROR: Result := 'Unknown error occured.';
        APEV2LIBRARY_ERROR_NO_TAG_FOUND: Result := 'No APE(v2) tag found.';
        APEV2LIBRARY_ERROR_EMPTY_TAG: Result := 'APE(v2) tag is empty.';
        APEV2LIBRARY_ERROR_EMPTY_FRAMES: Result := 'APE(v2) tag contains only empty frames.';
        APEV2LIBRARY_ERROR_OPENING_FILE: Result := 'Error opening file.';
        APEV2LIBRARY_ERROR_READING_FILE: Result := 'Error reading file.';
        APEV2LIBRARY_ERROR_WRITING_FILE: Result := 'Error writing file.';
        APEV2LIBRARY_ERROR_CORRUPT: Result := 'Error: corrupt file.';
        APEV2LIBRARY_ERROR_NOT_SUPPORTED_VERSION: Result := 'Error: not supported APE tag version.';
        APEV2LIBRARY_ERROR_NOT_SUPPORTED_FORMAT: Result := 'Error: not supported file format.';
    end;
end;

Initialization

    APEv2ID[1] := Ord('A');
    APEv2ID[2] := Ord('P');
    APEv2ID[3] := Ord('E');
    APEv2ID[4] := Ord('T');
    APEv2ID[5] := Ord('A');
    APEv2ID[6] := Ord('G');
    APEv2ID[7] := Ord('E');
    APEv2ID[8] := Ord('X');

    ID3v1ID[0] := Ord('T');
    ID3v1ID[1] := Ord('A');
    ID3v1ID[2] := Ord('G');

    ID3LYRICSTAGIDSTART[1] := Ord('L');
    ID3LYRICSTAGIDSTART[2] := Ord('Y');
    ID3LYRICSTAGIDSTART[3] := Ord('R');
    ID3LYRICSTAGIDSTART[4] := Ord('I');
    ID3LYRICSTAGIDSTART[5] := Ord('C');
    ID3LYRICSTAGIDSTART[6] := Ord('S');
    ID3LYRICSTAGIDSTART[7] := Ord('B');
    ID3LYRICSTAGIDSTART[8] := Ord('E');
    ID3LYRICSTAGIDSTART[9] := Ord('G');
    ID3LYRICSTAGIDSTART[10] := Ord('I');
    ID3LYRICSTAGIDSTART[11] := Ord('N');

    ID3LYRICSTAGIDEND[1] := Ord('L');
    ID3LYRICSTAGIDEND[2] := Ord('Y');
    ID3LYRICSTAGIDEND[3] := Ord('R');
    ID3LYRICSTAGIDEND[4] := Ord('I');
    ID3LYRICSTAGIDEND[5] := Ord('C');
    ID3LYRICSTAGIDEND[6] := Ord('S');
    ID3LYRICSTAGIDEND[7] := Ord('2');
    ID3LYRICSTAGIDEND[8] := Ord('0');
    ID3LYRICSTAGIDEND[9] := Ord('0');

end.


